
#define F_CPU 16000000UL

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lcd_lib.h"

#define Reset_State		0
#define Counting_State	1
#define Print_State		2

#define SW1_PRESSED (~PINB & (1<<PINB1))
#define SW2_PRESSED (~PINB & (1<<PINB7))


// Useful variables
volatile char DebounceFlag1;
volatile char DebounceFlag2;
volatile char update_LCD;

volatile uint16_t timer1_overflows_snapshot;
volatile uint16_t last_T1_value;
uint16_t timer1_overflows;

volatile char StopWatch_State;
volatile unsigned long total_time;
volatile unsigned long lap_time;
volatile unsigned long best_lap_time;

char lcd_buffer[17];	// LCD display buffer
const uint8_t LCD_Ready[] PROGMEM = "Ready!";
const uint8_t LCD_Counting[] PROGMEM = "Counting...";
const uint8_t LCD_Total_Time[] PROGMEM = "Total: ";
const uint8_t LCD_Lap_Time[] PROGMEM = "Lap:            ";
const uint8_t LCD_Best_Lap_Time[] PROGMEM = "Best: ";

//-------------------------------------------------------------------------------
void start_Timer0()
{
	// Set up Timer0 in Normal mode. COMPA Interrupt after 4ms at 16MHz/1024 Clock frequency
	TCCR0B = (1<<CS02) | (1<<CS00);		// Set pre-scalar to divide by 1024
}

void stop_Timer0()
{
	TCCR0B &= ~((1<<CS02) | (1<<CS01) | (1<<CS00));		// Clear only the prescaler bits
	TCNT0 = 0;		// Reset the counter
}

void start_Timer1()
{
	TCNT1 = 0;
	// Set up Timer1 in Normal mode to overflow after 4.194304s at 16MHz/1024 Clock frequency
	TCCR1B |= (1<<CS12) | (1<<CS10);		// Set pre-scalar to divide by 1024. Each tick is 64 microseconds
}

void stop_Timer1()
{
	TCCR1B &= ~((1<<CS12) | (1<<CS11) | (1<<CS10));		// Clear only the prescaler bits
	TCNT1 = 0;		// Reset the counter
}

//-------------------------------------------------------------------------------
// All initializations
void initialize_all(void)
{
	// start the LCD
	initialize_LCD();
	LCDcursorOFF();
	LCDclr();
	CopyStringtoLCD(LCD_Ready, 0, 0);
	
	//Setting Ports.
	DDRD |= (1<<DDD7)|(1<<DDD6);	// LED1 & LED2 are outputs
	PORTD = 0x00;					// Initially LEDs are OFF
	
	// Initializing variables
	DebounceFlag1 = 0;
	DebounceFlag2 = 0;
	update_LCD = 0;
	
	StopWatch_State = Reset_State;
	timer1_overflows = 0;
	timer1_overflows_snapshot = 0;
	last_T1_value = 0;
	total_time = 0;
	lap_time = 0;
	best_lap_time = 0xFFFF;
	
	// Pin Change Interrupt Configuration //
	DDRB &= ~((1<<DDB7)|(1<<DDB1));			// PB7 (PCINT7) and PB1 (PCINT7) Switches are inputs
	PCICR |= (1<<PCIE0);					// Enable Pin Change Interrupt 0
	PCMSK0 |= (1<<PCINT7)|(1<<PCINT1);		// Enable pins 1 and 7 in the PCINT mask
	
	// Set up Timer0 in CTC mode to overflow after 4ms at 16MHz/256 Clock frequency
	TCCR0A |= (1<<WGM01);				// CTC Mode
	OCR0A  = 250;  						// Set the compare reg to 250 time ticks
	TIMSK0 |= (1<<OCIE0A);				// turn on Timer0 Compare match ISR
	
	// Timer 1
	TIMSK1 |= (1<<TOIE1);			// Enable Overflow Interrupt.
}

//-------------------------------------------------------------------------------
// Timer 1 overflow ISR
ISR (TIMER1_OVF_vect)
{
	// Increment the Stopwatch time counter
	timer1_overflows++;
}

//-------------------------------------------------------------------------------
// Pin Change Interrupt 0 ISR
ISR(PCINT0_vect)
{
	// Disable Pin Change Interrupt
	PCICR &= ~(1<<PCIE0);
	
	// Record this event
	switch (StopWatch_State)
	{
		case Reset_State:
		start_Timer1();
		break;
		
		case Counting_State:
		timer1_overflows_snapshot = timer1_overflows;
		last_T1_value = TCNT1;	// copy the recorded snapshot
		break;
		
		case Print_State:
		break;
	}
	
	// Set flag to enable button debouncing
	if(SW1_PRESSED)
		DebounceFlag1 = 1;
	else if(SW2_PRESSED)
		DebounceFlag2 = 1;
	
	// Start Timer0 to for Debounce Delay
	start_Timer0();
}

//-------------------------------------------------------------------------------
// Timer0 Compare Match A ISR (TCNT0 = OCR0A)
ISR (TIMER0_COMPA_vect)
{
	// Stop Timer0
	stop_Timer0();
	
	// Take action
	switch (StopWatch_State)
	{
		case Reset_State:
		if(DebounceFlag1 && SW1_PRESSED)	// Start of race
		{
			// Turn on LED1
			PORTD |= (1<<PORTD7
			update_LCD = 1;
			StopWatch_State = Counting_State;
		}
		
		case Counting_State:
		if(DebounceFlag1 && SW1_PRESSED)	// End of race
		{
			// Turn off LEDs
			PORTD &= ~((1<<PORTD7)|(1<<PORTD6));
			stop_Timer1();
			DebounceFlag1 = 0;
			lap_time = (timer1_overflows_snapshot * 4194.304) + (last_T1_value * 0.064) - total_time;
			total_time += lap_time;
			if(lap_time < best_lap_time)
			{
				best_lap_time = lap_time;
			}
			update_LCD = 1;
			StopWatch_State = Print_State;
		}
		else if(DebounceFlag2 && SW2_PRESSED)	// Record the Lap time
		{
			// Toggle LED2
			PORTD ^= (1<<PORTD6);
			DebounceFlag2 = 0;
			lap_time = (timer1_overflows_snapshot * 4194.304) + (last_T1_value * 0.064) - total_time;
			total_time += lap_time;
			if(lap_time < best_lap_time)
			{
				best_lap_time = lap_time;
			}
			update_LCD = 1;
		}
		break;
		
		case Print_State:
		if(DebounceFlag1 && SW1_PRESSED)
		{
			// Reset everything 
			total_time = 0;
			lap_time = 0;
			best_lap_time = 0xFFFF;
			last_T1_value = 0;					// Reset last timer ticks
			timer1_overflows = 0;				// Reset Overflow counts
			timer1_overflows_snapshot = 0;		// Reset snapshot
			TCNT1 = 0;
			
			update_LCD = 1;
			StopWatch_State = Reset_State;
		}
		break;
	}
	
}

//-------------------------------------------------------------------------------
// Displays Total Time on LCD
void display_time(unsigned long time, uint8_t row)
{	
	sprintf(lcd_buffer, "%lu ms", time);
	
	// Print on LCD
	if(row == 0)
	{
		CopyStringtoLCD(LCD_Total_Time, 0, 0);
		LCDGotoXY(7,0);
		LCDstring((uint8_t*) lcd_buffer, strlen(lcd_buffer));
	}
	else if(row == 1)
	{
		CopyStringtoLCD(LCD_Lap_Time, 0, 1);
		LCDGotoXY(7,1);
		LCDstring((uint8_t*) lcd_buffer, strlen(lcd_buffer));
	}
	else
	{
		CopyStringtoLCD(LCD_Best_Lap_Time, 0, 1);
		LCDGotoXY(7,1);
		LCDstring((uint8_t*) lcd_buffer, strlen(lcd_buffer));
	}
}


//-------------------------------------------------------------------------------
// Button Debounce task
void refresh_LCD(void)
{
	update_LCD = 0;

	// Update the Stopwatch State
	switch (StopWatch_State)
	{
		case Reset_State:
		LCDclr();
		CopyStringtoLCD(LCD_Ready, 0, 0);
		break;
		
		case Counting_State:
		CopyStringtoLCD(LCD_Counting, 0, 0);
		if(lap_time != 0)
			display_time(lap_time, 1);			// Print the lap time
		break;
		
		case Print_State:
		LCDclr();
		display_time(total_time, 0);
		display_time(best_lap_time, 2);
		break;
	}		
}

//-------------------------------------------------------------------------------
int main(void)
{	
	// Initialize everything
	initialize_all();
	// Enable Global Interrupts
	sei();

	// ------ Event loop ------ //
	while (1) 
	{
		if(update_LCD)
		{
			refresh_LCD();
		}
		
	} // End event loop
}
